/*
 *  Copyright (c) 2017 Intel Corporation. All Rights Reserved.
 *
 *  Portions of the source code contained or described herein and all documents related
 *  to portions of the source code ("Material") are owned by Intel Corporation or its
 *  suppliers or licensors.  Title to the Material remains with Intel
 *  Corporation or its suppliers and licensors.  The Material contains trade
 *  secrets and proprietary and confidential information of Intel or its
 *  suppliers and licensors.  The Material is protected by worldwide copyright
 *  and trade secret laws and treaty provisions.  No part of the Material may
 *  be used, copied, reproduced, modified, published, uploaded, posted,
 *  transmitted, distributed, or disclosed in any way without Intel's prior
 *  express written permission.
 *
 *  No license under any patent, copyright, trade secret or other intellectual
 *  property right is granted to or conferred upon you by disclosure or
 *  delivery of the Materials, either expressly, by implication, inducement,
 *  estoppel or otherwise. Any license under such intellectual property rights
 *  must be express and approved by Intel in writing.
 */
#include <stdlib.h>
#include <stdio.h>
#include "omp.h"
#include "mkl.h"

#define min(x,y) (((x) < (y)) ? (x) : (y))

int offload(int m, int n, int k, double alpha, double beta)
{
   int i, j;

   /* Allocate memory using MKL function to aligned on 64-byte boundary */
   double *A = mkl_malloc(sizeof(double) * m * k, 64);
   if (A == NULL) 
      return (-1);
   else
   {
      /* Initialize matrix A */
      for (i = 0; i < m*k; i++)
         A[i] = 1.0;
   }

   double *B = mkl_malloc(sizeof(double) * k * n, 64);
   if (B == NULL) 
   {
      mkl_free(A);
      return (-1);
   }
   else
   {
      /* Initialize matrix B */
      for (i = 0; i < k*n; i++)
         B[i] = 2.0;
   }

   double *C = mkl_malloc(sizeof(double) * m * n, 64);
   if (C == NULL)
   {
      mkl_free(A);
      mkl_free(B);
      return (-1);
   }
   else
   {
      /* Initialize matrix C */
      for (i = 0; i < m*n; i++)
         C[i] = 0.0;
   }

   const int NITERS = 3;

   for (i = 0; i < NITERS; i++) 
   {
      static int first_run = 1, last_run = 0;

#pragma offload target(mic:0) in(m, n, k, alpha, beta) \
		in(A: length(m*k) alloc_if(first_run) free_if(last_run)) \
		in(B: length(k*n) alloc_if(first_run) free_if(last_run)) \
		inout(C: length(m*n) alloc_if(first_run) free_if(last_run))
      {
         cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
	   m, n, k, alpha, A, k, B, n, beta, C, n);
      }

   
      first_run = 0;
      if (i == NITERS-2)
         last_run = 1;
   }

   // Verify
   printf (" Top left corner of matrix A(m x k): \n");
   for (i=0; i<min(m,6); i++) 
   {
      for (j=0; j<min(k,6); j++) 
         printf ("%12.2f", A[j+i*k]);

      printf ("\n");
   }
        
   printf ("\n Top left corner of matrix B(k x n): \n");
   for (i=0; i<min(k,6); i++) 
   {
      for (j=0; j<min(n,6); j++)
         printf ("%12.2f", B[j+i*n]);
 
      printf ("\n");
   }
    
   printf ("\n Top left corner of matrix C(m x n): \n");
   for (i=0; i<min(m,6); i++) 
   {
      for (j=0; j<min(n,6); j++)
         printf ("%12.2f", C[j+i*n]);

      printf ("\n");
   }

   mkl_free(A);
   mkl_free(B);
   mkl_free(C);

   return 0;
}

int main(int argc, char **argv)
{
   int rc, m, n, k;
   double alpha, beta;

   if (argc != 6) 
   {
      printf("Usage: ./mkl-cao.out m n k alpha beta \n");
      printf("Where m is the number of row of matrix A \n");
      printf("      n is the number of column of matrix B \n");
      printf("      k is the number of column of matrix A \n");
      printf("      alpha and beta are scale factors  \n");

      return argc;
   }

   m = atoi(argv[1]);
   n = atoi(argv[2]);
   k = atoi(argv[3]);
   alpha = atof(argv[4]);
   beta = atof(argv[5]);

   printf("m:%d n:%d k:%d alpha:%8.2f beta:%8.2f\n", m, n, k, alpha, beta);
   rc = offload(m, n, k, alpha, beta);

   return rc;
}

